package demux;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import muxer.MuxManager;

import common.AudioStreamDescriptor;
import common.ExternalProcess;
import common.FileStream;
import common.InputStreamBuffer;
import common.Job;
import common.JobStreamI;
import common.StreamDescriptor;
import common.VideoStreamDescriptor;
import common.StreamIdentifier.DolbyDigitalAnalzyer;
import common.StreamIdentifier.DolbyTrueHdAnalyzer;
import common.StreamIdentifier.DtsAnalyzer;
import common.StreamIdentifier.DtsMasterAudioAnalyzer;
import common.StreamIdentifier.LpcmAnalyzer;
import common.StreamIdentifier.MpegVideoAnalyzer;
import common.StreamIdentifier.VC1VideoAnalyzer;
import common.avc.AvcConstants;

public class Eac3toDemuxer extends Demuxer {
	private static final String VIDEO[][]={
		{"VC-1",VC1VideoAnalyzer.CODEC_NAME, VC1VideoAnalyzer.FILE_EXTENSION},
		{"h264/AVC", AvcConstants.CODEC_NAME, AvcConstants.FILE_EXTENSION},
		{"MPEG2", MpegVideoAnalyzer.CODEC_NAME, MpegVideoAnalyzer.FILE_EXTENSION}
	};
	private static final String AUDIO[][]={
		{"E-AC3", DolbyDigitalAnalzyer.EAC3_CODEC_NAME, DolbyDigitalAnalzyer.EAC3_FILE_EXTENSION},
		{"AC3", DolbyDigitalAnalzyer.AC3_CODEC_NAME, DolbyDigitalAnalzyer.AC3_FILE_EXTENSION},
		{"TrueHD", DolbyTrueHdAnalyzer.CODEC_NAME, DolbyTrueHdAnalyzer.FILE_EXTENSION},
		{"DTS Master Audio", DtsMasterAudioAnalyzer.CODEC_NAME, DtsMasterAudioAnalyzer.CODEC_NAME},
		{"RAW/PCM", LpcmAnalyzer.CODEC_NAME, ".wav"}, 
		{"DTS", DtsAnalyzer.CODEC_NAME, DtsAnalyzer.FILE_EXTENSION}
	};
	
	private static final String SUBTITLE="Subtitle";
	private static final String CHAPTERS="Chapters";
	
	private static final String CHANNELS="channels";
	private static final String KBPS="kbps";
	private static final String KHZ="khz";
	private static final String DELAY="ms";
	private String eac3toPath;
	
	public Eac3toDemuxer(String eac3toPath) {
		this.eac3toPath=eac3toPath;
	}
	@Override
	public Map<Integer, StreamDescriptor> getStreamDescriptorMap(
			JobStreamI stream, MuxManager muxManager) throws IOException {
		FileStream diskStream=(FileStream)stream;
		File files[]=diskStream.getFiles();
		StringBuffer sb=new StringBuffer();
		String sep="";
		for (int i=0;i<files.length;i++) {
			sb.append(sep);
			sep="+";
			sb.append(files[i].getAbsolutePath());
		}
		
		//ProcessDialog processDialog=new ProcessDialog();
		stream.getJob().setSize(300*Job.KB);
		InputStreamBuffer logBuffer=new InputStreamBuffer();

		ExternalProcess process=new ExternalProcess(stream.getJob(),"analyze: ", null, logBuffer);
		int rc=process.start(new String[]{new File(eac3toPath).getAbsolutePath(),sb.toString(),"-progressNumbers"});
		String log=logBuffer.toString();
		if (rc==0) {
			logger.info(log);
		} else {
			logger.error(log);
			throw new IOException("Analyze Streams Failed");
		}
		BufferedReader reader=new BufferedReader(new StringReader(log));
		ArrayList<String> lineList=new ArrayList<String>(30);
		for (String line;(line=reader.readLine())!=null;) {
			lineList.add(line.trim());
		}
		HashMap<Integer,StreamDescriptor> streamDescriptorMap=new HashMap<Integer,StreamDescriptor>();
		//Start at 1 to skip c:\...\eac3to
		for (int i=1;i<lineList.size();i++) {
			String line=lineList.get(i);
			int colonPos=line.indexOf(':');
			if (colonPos==1 || colonPos==2) {
				Integer index=Integer.parseInt(line.substring(0, colonPos));
				String parts[]=line.substring(colonPos+2).split(",");
				if (parts[0].startsWith(SUBTITLE)) {
					StreamDescriptor sd=new StreamDescriptor(index.intValue(),StreamDescriptor.CLASSIFICATION_SUBTITLE);
					sd.setCodec(parts[0]);
					if (parts.length > 1) {
						sd.setLanguage(parts[1].trim());
					}
					streamDescriptorMap.put(index, sd);
				} else if (parts[0].startsWith(CHAPTERS)) {
					StreamDescriptor sd=new StreamDescriptor(index.intValue(),StreamDescriptor.CLASSIFICATION_EAC3TO_CHAPTERS);
					sd.setCodec(parts[0]);
					if (parts.length > 1) {
						sd.setLanguage(parts[1].trim());
					}
					streamDescriptorMap.put(index, sd);
				} else {
					StreamDescriptor sd=null;
					for (int a=0;a<AUDIO.length;a++) {
						if (parts[0].startsWith(AUDIO[a][0])) {
							AudioStreamDescriptor ad=new AudioStreamDescriptor(index.intValue());
							sd=ad;
							ad.setCodec(AUDIO[a][1]);
							ad.setFileNameExtension(AUDIO[a][2]);
							for (int p=1;p<parts.length;p++) {
								String part=parts[p].trim();
								if (part.endsWith(CHANNELS)) {
									ad.setChannels(part.substring(0,3));
								} else if (part.endsWith(KBPS)) {
									ad.setBitRate(Integer.parseInt(part.substring(0, part.indexOf(KBPS))));
								} else if (part.endsWith(KHZ)) {
									ad.setFrequency(Integer.parseInt(part.substring(0, part.indexOf(KHZ)))*1000);
								} else if (part.endsWith(DELAY)) {
									ad.setDelay(Integer.parseInt(part.substring(0, part.indexOf(DELAY))));
								} else if (p==1) {
									ad.setLanguage(part);
								}
								
							}
							if (i<lineList.size()-1) {
								String descript=lineList.get(i+1);
								if (descript.charAt(0)=='"') {
									sd.setLanguage(sd.getLanguage() + " " + descript);
								}
							}
							break;
						}
					}
					if (sd==null) {
						for (int v=0;v<VIDEO.length;v++) {
							if (parts[0].startsWith(VIDEO[v][0])) {
								VideoStreamDescriptor vd=new VideoStreamDescriptor(index.intValue());
								sd=vd;
								vd.setCodec(VIDEO[v][1]);
								vd.setFileNameExtension(VIDEO[v][2]);
								String videoDescriptor=parts[1].trim();
								StringBuffer res=new StringBuffer(4);
								int r=0;
								while (videoDescriptor.charAt(r) <='9') {
									res.append(videoDescriptor.charAt(r++));
								}
								vd.setHeight(Integer.parseInt(res.toString()));
								int aspectStart=videoDescriptor.indexOf('(');
								if (aspectStart > 0) {
									int aspectEnd=videoDescriptor.indexOf(')', aspectStart);
									String aspectString=videoDescriptor.substring(aspectStart + 1, aspectEnd);
									String aspectStringArray[]=aspectString.split(":");
									int aspectArray[]={Integer.parseInt(aspectStringArray[0]),Integer.parseInt(aspectStringArray[1])};
									vd.setAspectRatio(aspectArray);
									vd.setWidth(vd.getHeight()*aspectArray[0]/aspectArray[1]);
								}
								if (videoDescriptor.charAt(r)=='i') {
									vd.setInterlaced(1);
								} else if (videoDescriptor.charAt(r)=='p') {
									vd.setInterlaced(0);
								}
								break;
							}
						}
					}
					if (sd != null) {
						streamDescriptorMap.put(index, sd);
					}
				}
			}
		}
		return streamDescriptorMap;
	}

	@Override
	public void parse(JobStreamI stream, MuxManager muxManager)
			throws IOException {
		throw new RuntimeException("Eac3toDemuxer is only used for identifying streams");

	}

}
