package demux.pes;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.WritableByteChannel;

import common.AvcVideoAnalyzer;
import common.StreamDescriptor;
import common.StreamIdentifier;
import common.Util;
import common.mpeg.MpegConstants;
import common.mpeg.MpegUtil;
import common.packet.PacketPart;
import common.packet.PesPacket;
import common.packet.PesPacketPart;

public class PesUtil implements MpegConstants {
	public static final int MAX_PACKET_LEN=256*256-1;
	private int packetSize=0x07EC+6;
	private byte[] emptyHeader=new byte[MpegConstants.EMPTY_PES_HEADER.length];
	
	/**
	 * Build a new PesPacketHeader and set the stream type and pts
	 * @param sd
	 * @param pts if < 0 do nothing
	 * @return
	 */
	public static PesPacketPart getPesHeader(StreamDescriptor sd, long pts) {
		byte emptyHeader[]= pts > 0 ? PTS_PES_HEADER : EMPTY_PES_HEADER;
		ByteBuffer buffer=ByteBuffer.allocateDirect(emptyHeader.length + 4);
		buffer.put(emptyHeader);
		buffer.limit(emptyHeader.length);
		PesPacketPart header=new PesPacketPart(buffer);
		if (pts > 0) {
			MpegUtil.setPesTime(buffer, 9, pts);
		}
		
		byte type;
		if (sd != null) {
			String codec=sd.getCodec();
			if (StreamIdentifier.MpegVideoAnalyzer.CODEC_NAME.equals(codec) || AvcVideoAnalyzer.CODEC_NAME.equals(codec)) {
				type=VIDEO_PACKET;
			} else if (StreamIdentifier.DolbyDigitalAnalzyer.AC3_CODEC_NAME.equals(codec)) {
				type=PRIVATE_PACKET;
			} else if (codec.startsWith(StreamIdentifier.MpegAudioAnalyzer.CODEC_NAME)) { 
				type=AUDIO_PACKET;
			} else if (codec.startsWith(StreamIdentifier.AdvancedAudioCodecAnalyzer.CODEC_NAME)) { 
				type=AUDIO_PACKET_LATM;
			} else {
				type=PRIVATE_PACKET;
			}
		} else {
			type=PRIVATE_PACKET;
		}
		buffer.put(3, type);
		return header;
	}
	
	/**
	 * Primarily Used for AC3 packets, but applies any last minute changes
	 * Adds the AC3 header stream
	 * @param packet
	 */
	public static void finalizePrivateStreamHeader(PesPacket packet, StreamDescriptor sd, Byte subStream) {
		//bytes >= 0x80 are negative
		if (subStream.byteValue() < 0) {
			int count=0;
			int offset=1;
			byte sync[];
			if (StreamIdentifier.DolbyDigitalAnalzyer.AC3_CODEC_NAME.equals(sd.getCodec())) {
				sync=StreamIdentifier.DolbyDigitalAnalzyer.SYNC;
			} else if (StreamIdentifier.DtsAnalyzer.CODEC_NAME.equals(sd.getCodec())) {
				sync=StreamIdentifier.DtsAnalyzer.SYNC;
			} else {
				return;
			}
			for (int i=0; (i=Util.findBytes(packet.getPayload().getByteBuffer(), sync, i)) >= 0; i++) {
				if (count==0) {
					offset=i+1;
				}
				count++;
			}
			byte psHeader[]={subStream, (byte)count, 0, 0};
			Util.set16(psHeader, 2, offset);
			packet.getHeader().appendFrom(psHeader, 0, psHeader.length);
		} else {
			packet.getHeader().put(subStream);
		}
	}
	
	public PesUtil() {
		System.arraycopy(MpegConstants.EMPTY_PES_HEADER, 0, emptyHeader, 0, emptyHeader.length);
	}
	final public void writePacket(final PesPacket packet, final WritableByteChannel out) throws IOException {
		PesPacketPart header=packet.getHeader();
		final PacketPart payload=packet.getPayload();
		int bytesRemaining=payload.limit();
		
		final PesPacketPart emptyHeader=(bytesRemaining + header.limit() > packetSize) ? getEmptyHeader(header) : null;
		
		for (int offset=0;bytesRemaining > 0;) {
			int writeSize=Math.min(bytesRemaining, packetSize-header.limit());
			header.set16(writeSize + header.limit() -6,4);
			header.write(out);
			packet.getPayload().write(out, offset, writeSize);
			offset+=writeSize;
			bytesRemaining-=writeSize;
			header=emptyHeader;
		}
	}

	public int getPacketSize() {
		return packetSize;
	}

	public void setPacketSize(int packetSize) {
		this.packetSize = packetSize;
	}
	/**
	 * Return a PES header with the PTS and DTS cleared
	 * @param header
	 * @return
	 */
	private PesPacketPart getEmptyHeader(PesPacketPart header) {
		byte ptsDtsFlag=header.get(7);
		int skipBytes=0;
		if ((ptsDtsFlag & MpegUtil.PTS_PRESENT)==MpegUtil.PTS_PRESENT) {
			skipBytes+=5;
		}
		if ((ptsDtsFlag & MpegUtil.DTS_PRESENT)==MpegUtil.DTS_PRESENT) {
			skipBytes+=5;
		}
		final int emptyLen=header.limit()-skipBytes;
		PesPacketPart emptyHeader=new PesPacketPart(emptyLen);
		emptyHeader.appendFrom(header.getByteBuffer(), 0, 9);
		emptyHeader.appendFrom(header.getByteBuffer(), 9+skipBytes, emptyLen-9);
		//remove the PTS/DTS flags from the header
		ptsDtsFlag=(byte)(ptsDtsFlag & (~(MpegUtil.PTS_PRESENT + MpegUtil.DTS_PRESENT)));
		emptyHeader.setByte(ptsDtsFlag, 7);
		//Remove the PTS/DTS bytes from the header length
		emptyHeader.setByte((byte)((header.get(8) & 0xff)-skipBytes), 8);
		return emptyHeader;
	}
}
