package muxer.pes;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.channels.WritableByteChannel;
import java.util.HashMap;
import java.util.Map;

import muxer.MuxFilter;
import muxer.MuxManager;
import muxer.MuxUtil;
import muxer.PacketBuffer;
import regulator.StreamPtsPacketCountRegulator;

import common.FileStream;
import common.Job;
import common.Logger;
import common.StreamDescriptor;
import common.mpeg.MpegConstants;
import common.mpeg.MpegUtil;
import common.packet.ClockReferenceI;
import common.packet.Packet;
import common.packet.PacketList;
import common.packet.PesPacket;
import common.packet.PesPacketPart;

import demux.pes.PesUtil;
import demux.ty.TyPesDemuxer;

public class PesMuxer extends MuxFilter implements MpegConstants {
	private final static long MIN_SCR_SYNC_STREAM_OFFSET=PTS_FREQ/100; //10 millis
	
	private Logger logger=Logger.getLogger(this.getClass());
	private PacketList packetList=new PacketList();
	private int syncStreamId=-1;
	private long syncDelta=Long.MIN_VALUE;
	private long lastScr=Long.MIN_VALUE;
	private long leadTime;
	private PesUtil pesUtil=new PesUtil();
	private WritableByteChannel out;
	private HashMap<Integer, Byte> privateStreamMap=new HashMap<Integer, Byte>();
	
	public PesMuxer(WritableByteChannel out) {
		this.out=out;
	}
	protected void init() throws IOException {
		Map<Integer, StreamDescriptor> muxedStreamDescriptorMap=muxManager.getMuxedStreamDescriptorMap();
		syncStreamId=MuxUtil.getLinearStream(muxedStreamDescriptorMap).getStreamId();
		PacketBuffer packetBuffer=muxManager.getPacketBuffer();
		StreamPtsPacketCountRegulator regulator=new StreamPtsPacketCountRegulator(new int[]{syncStreamId}, 6);
		if (packetBuffer==null) {
			muxManager.add(packetBuffer=new PacketBuffer());
		}
		packetBuffer.addRegulator(regulator);
		
		int audioStream=0x80;
		int subtitleStream=0x20;
		for (StreamDescriptor streamDescriptor : muxedStreamDescriptorMap.values()) {
			//StreamDescriptor streamDescriptor=(StreamDescriptor)i.next();
			if (streamDescriptor.getClassification()==StreamDescriptor.CLASSIFICATION_AUDIO) {
				if (streamDescriptor.getTsType() != TS_STREAM_TYPE_MPEG1_AUDIO && streamDescriptor.getTsType() != TS_STREAM_TYPE_MPEG2_AUDIO) {
					privateStreamMap.put(streamDescriptor.getStreamId(), (byte)audioStream++);
				}
			} else if (streamDescriptor.getClassification()==StreamDescriptor.CLASSIFICATION_SUBTITLE) {
				privateStreamMap.put(streamDescriptor.getStreamId(),(byte)subtitleStream++);
			}
		}
	}
	public int addPacket(Packet packet) throws IOException {
		packetList.add(packet);
		for (int i=0, syncCount=0;i<packetList.size();i++) {
			if (packetList.get(i).getStreamId()==syncStreamId) {
				syncCount++;
				if (syncCount > 1) {
					mux(i);
				}
			}
		}
		return MuxManager.CONTINUE;
	}
	public void streamEnd() throws IOException {
		if (!packetList.isEmpty()) {
			mux(packetList.size());
		}
		out.close();
		super.streamEnd();
	}
	public void mux(int lastPacket) throws IOException {
		long scr;
		if (syncDelta==Long.MIN_VALUE) {
			syncDelta=MuxUtil.getMode(muxManager.getPacketBuffer().iterator(), syncStreamId);
			long data[]=MuxUtil.calculateStreamTime(packetList, syncStreamId, syncDelta);
			scr=data[0];
			lastScr=scr-syncDelta;//prevents a warning about a jump in the stream
			leadTime=data[1]<MIN_SCR_SYNC_STREAM_OFFSET? MIN_SCR_SYNC_STREAM_OFFSET : data[1];
			writePackHeader(lastScr, MUX_RATE, out);
			
			
			MpegUtil.writeSystemHeader(muxManager.getMuxedStreamDescriptorMap().values(), out);
			writePackHeader(scr, MUX_RATE, out);
		} else {
			Packet packet=packetList.get(0);
			scr=packet.getClockReference(ClockReferenceI.CLOCK_REFERENCE_TYPE_PTS);
			if (scr != ClockReferenceI.CLOCK_REFERENCE_NONE) {
				writePackHeader(scr - leadTime, MUX_RATE, out);
			}
		}
		for (int i=0;i<lastPacket;i++) {
			PesPacket packet=(PesPacket)packetList.remove(0);
			PesPacketPart header=packet.getHeader();
			//PacketPart payload=packet.getPayload();
			byte type=header.get(3);
			if (type==VC1_PACKET) {
				//Deal with the fact the some BluRay audio is in VC_1 style packets
				Map<Integer, StreamDescriptor> streamDescriptorMap=muxManager.getStreamDescriptorMap(null);
				StreamDescriptor sd=(StreamDescriptor)streamDescriptorMap.get(packet.getStreamId());
				if (sd.getClassification()==StreamDescriptor.CLASSIFICATION_AUDIO) {
					type=PRIVATE_PACKET;
					header.setByte(type, 3);
				}
			}
			if (type==PRIVATE_PACKET) {
				Byte subStream=(Byte)privateStreamMap.get(Integer.valueOf(packet.getStreamId()));
				header.put(subStream.byteValue());
				//bytes > 0x80 are negative
				if (subStream.byteValue() < 0) {
					header.appendFrom(PRIVATE_SUBSTREAM_HEADER, 1, PRIVATE_SUBSTREAM_HEADER.length-1);
				}
				//Don't split private packets.  This should be fixed
				header.set16(packet.getPayload().limit() + header.limit() -6,4);
				header.write(out);
				packet.getPayload().write(out);
			} else {
				pesUtil.writePacket(packet, out);
			}
		}
	}
	protected void writePackHeader(long systemClock, int muxRate, WritableByteChannel out) throws IOException {
		if (systemClock < 0) {
			logger.error("Attempted to write negative SCR " + systemClock);
			systemClock=0;
		}
		if (lastScr !=0) {
			if (systemClock == 0 || systemClock < lastScr) {
				//Hack
				//Keep the clock moving forward
				//systemClock = lastScr+Util.calculateFrameTime(videoPacket.getFrameRateType());
				logger.warn("Attempted to move the SCR back " + (lastScr-systemClock));
				//lastScr+=100;
				systemClock=lastScr;
			}
			if (systemClock > lastScr + 1000000) {
				logger.warn("Huge SCR Jump!  Delta: " + (systemClock - lastScr) + "  New SCR:" + MpegUtil.pesTimeToString(systemClock) + "  Last SCR: " + MpegUtil.pesTimeToString(lastScr));
			}
		}
		
		
		lastScr=systemClock;
		MpegUtil.setPackHeader(PACK_HEADER, 4, systemClock, muxRate);
		out.write(PACK_HEADER);
		/*if (!wroteSystemHeader) {
			wroteSystemHeader=true;
			writeSystemHeader(out, systemClock);
		}*/
	}
	
	public static void main(String args[]) {
		try {
			//String fileName="The Magic School Bus - The Magic School Bus Goes on Air";
			String fileName="CSI_ Crime Scene Investigation - Monster in the Box";
			System.out.println("Starting");
			String titleName="F:\\test Tys\\"+fileName+".ty";
			FileStream fileJob=new FileStream(new Job(null), new File[]{new File(titleName)});
			TyPesDemuxer demuxer=new TyPesDemuxer();
			//new RestreamingPesMuxer(buffer, new NullOutputStream());
			long start=System.currentTimeMillis();
			//in.skip(10631512064l);
			MuxManager muxManager=new MuxManager(demuxer);
			Map<Integer, StreamDescriptor> map=muxManager.getStreamDescriptorMap(fileJob);
			System.out.println(map);
			//PacketBuffer packetBuffer=muxManager.getPacketBuffer();
			//packetBuffer.setRegulator(PacketBuffer.ALWAYS_FULL);
			//muxManager.remove(packetBuffer);
			muxManager.add(new PesMuxer(new FileOutputStream("C:/"+fileName+".mpg").getChannel()));
			muxManager.mux(fileJob);
			System.out.println("Done: " + (System.currentTimeMillis()-start));
		} catch (IOException e) {
			System.out.println(e);
		}
	}
	public int getPosition() {
		return POSITION_MUXER;
	}	
}
